@@ -9,8 +9,11 @@
 from __future__ import annotations
 
 import logging
+import os
 
-import clu.parser
+from typing import List, Optional, Union
+
+import clu.parsers.click
 from clu import BaseActor, JSONActor
 from clu.tools import ActorHandler
 from sdsstools.logger import SDSSLogger
@@ -24,7 +27,7 @@ from . import commands
 __all__ = ["BaseCameraActor", "CameraActor"]
 
 
-class BaseCameraActor:
+class BaseCameraActor(BaseActor):
     """Base class for a camera CLU-like actor class.
 
     Expands a `CLU <https://clu.readthedocs.io/en/latest/>`__ actor
@@ -46,19 +49,22 @@ class BaseCameraActor:
         default in most command.
     command_parser
         The list of commands to use. It must be a command group deriving from
-        `~clu.parser.CluGroup` containing all the commands to use. If
+        `~clu.parsers.click.CluGroup` containing all the commands to use. If
         ``commands=None``, uses the internal command set.
+    schema
+        The path to the JSONSchema file with the actor datamodel. If ``"internal"``,
+        uses the default ``basecam`` model; `None` disables model validation.
     args,kwars
         Arguments and keyword arguments to be passed to the actor class.
-
     """
 
     def __init__(
         self,
         camera_system: CameraSystem,
         *args,
-        default_cameras: list[str] | str | None = None,
-        command_parser: clu.parser.CluGroup = None,
+        default_cameras: Union[List[str], str, None] = None,
+        command_parser: clu.parsers.click.CluGroup = None,
+        schema: Optional[str] = "internal",
         **kwargs,
     ):
 
@@ -67,6 +73,11 @@ class BaseCameraActor:
         assert camera_system is not None
         self.camera_system = camera_system
 
+        # Add actor to camera class context and all connected cameras.
+        self.camera_system.camera_class.fits_model.context.update({"__actor__": self})
+        for camera in self.camera_system.cameras:
+            camera.fits_model.context.update({"__actor__": self})
+
         #: An `.EventListener` that can be used to wait or respond to events.
         self.listener = EventListener()
         self.camera_system.notifier.register_listener(self.listener)
@@ -74,7 +85,10 @@ class BaseCameraActor:
         self.log: SDSSLogger
         self.parser = command_parser or commands.camera_parser
 
-        super().__init__(*args, **kwargs)
+        if schema == "internal":
+            schema = os.path.join(os.path.dirname(__file__), "schema.json")
+
+        super().__init__(*args, schema=schema, **kwargs)
 
         # Add commands that depend on what mixins the base camera has
         # been subclassed with.
@@ -103,6 +117,8 @@ class BaseCameraActor:
 
         # Check that at least one of the bases is a sublass of BaseActor
         for base in bases:
+            if base == BaseCameraActor:
+                continue
             if issubclass(base, BaseActor):
                 return
 
@@ -119,7 +135,7 @@ class BaseCameraActor:
                 for command in commands._MIXIN_TO_COMMANDS[mixin_name]:
                     self.parser.add_command(command)
 
-    def set_default_cameras(self, cameras: str | list[str] | None = None):
+    def set_default_cameras(self, cameras: Union[str, List[str], None] = None):
         """Sets the camera(s) that will be used by default.
 
         These cameras will be used by default when a command is issued without
