@@ -11,12 +11,24 @@ from __future__ import annotations
 import abc
 import asyncio
 import logging
+import os
 import pathlib
 import time
 import warnings
 from logging import INFO, WARNING
 
-from typing import Any, Callable, Generic, Optional, Type, TypeVar, Union, cast
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    List,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 
 import numpy
 
@@ -84,20 +96,20 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
         will be logged to the console.
     """
 
-    camera_class: Type[_T_BaseCamera] | None = None
+    camera_class: Union[Type[_T_BaseCamera], None] = None
     include = None
     exclude = None
 
     def __init__(
         self,
         camera_class: Optional[Type[_T_BaseCamera]] = None,
-        camera_config: Optional[AnyPath | dict[str, Any]] = None,
-        include: Optional[list[Any]] = None,
-        exclude: Optional[list[Any]] = None,
+        camera_config: Optional[Union[AnyPath, Dict[str, Any]]] = None,
+        include: Optional[List[Any]] = None,
+        exclude: Optional[List[Any]] = None,
         logger: Optional[SDSSLogger] = None,
         log_header: Optional[str] = None,
         log_file: Optional[AnyPath] = None,
-        verbose: Optional[bool | int] = False,
+        verbose: Optional[Union[bool, int]] = False,
     ):
 
         self.camera_class = camera_class or self.camera_class
@@ -118,7 +130,7 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
             self.logger.sh.setLevel(logging.WARNING)
 
         if log_file:
-            self.logger.start_file_logger(log_file)
+            self.logger.start_file_logger(str(log_file))
             if self.logger.fh:
                 self.logger.fh.formatter.converter = time.gmtime
                 self.log(f"logging to {log_file}")
@@ -127,20 +139,20 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
         self.loop.set_exception_handler(self.logger.asyncio_exception_handler)
 
         #: list: The list of cameras being handled.
-        self.cameras: list[_T_BaseCamera] = []
+        self.cameras: List[_T_BaseCamera] = []
 
         self._camera_poller = None
 
         #: .EventNotifier: Notifies of `.CameraSystemEvent` and `.CameraEvent` events.
         self.notifier = EventNotifier()
 
-        self._config: Optional[dict[str, Any]] = None
+        self._config: Optional[Dict[str, Any]] = None
 
         if camera_config:
             if isinstance(camera_config, dict):
                 self._config = camera_config.copy()
             else:
-                self._config = read_yaml_file(camera_config)
+                self._config = cast(Dict[str, Any], read_yaml_file(str(camera_config)))
                 self.log(f"read configuration file from {camera_config}")
 
         # If the config has a section named "cameras", prefer that.
@@ -151,20 +163,22 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
             if len(uids) != len(set(uids)):
                 raise ValueError("repeated UIDs in the configuration data.")
 
+        self.running: bool = False
+
     def setup(self):
         """Setup custom camera system.
 
         To be overridden by the subclass if needed. Must return ``self``.
-
         """
 
+        self.running = True
         return self
 
     def get_camera_config(
         self,
         name: Optional[str] = None,
         uid: Optional[str] = None,
-    ) -> dict[str, Any] | None:
+    ) -> Union[Dict[str, Any], None]:
         """Gets camera parameters from the configuration.
 
         Parameters
@@ -289,7 +303,7 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
                 await self.add_camera(uid=uid)
 
     @abc.abstractmethod
-    def list_available_cameras(self) -> list[str]:
+    def list_available_cameras(self) -> List[str]:
         """Lists the connected cameras as reported by the camera system.
 
         Returns
@@ -414,7 +428,7 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
         self,
         name: Optional[str] = None,
         uid: Optional[str] = None,
-    ) -> BaseCamera | Literal[False]:
+    ) -> Union[BaseCamera, Literal[False]]:
         """Gets a camera matching a name or UID.
 
         If only one camera is connected and the method is called without
@@ -487,6 +501,8 @@ class CameraSystem(LoggerMixIn, Generic[_T_BaseCamera], metaclass=abc.ABCMeta):
         if self._camera_poller and self._camera_poller.running:
             await self.stop_camera_poller()
 
+        self.running = False
+
     @property
     @abc.abstractmethod
     def __version__(self) -> str:
@@ -536,8 +552,8 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
         sequential counter.
     """
 
-    fits_model: FITSModel = basic_fits_model
-    image_namer: ImageNamer = ImageNamer(
+    fits_model = basic_fits_model
+    image_namer = ImageNamer(
         "{camera.name}-{num:04d}.fits",
         dirname=".",
         overwrite=False,
@@ -549,7 +565,7 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
         camera_system: CameraSystem,
         name: Optional[str] = None,
         force: bool = False,
-        image_namer: Optional[ImageNamer | dict] = None,
+        image_namer: Optional[Union[ImageNamer, dict]] = None,
         camera_params={},
     ):
 
@@ -616,18 +632,18 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
             self.connected = True
         except CameraConnectionError as ee:
             self.connected = False
-            self._notify(CameraEvent.CAMERA_CONNECT_FAILED)
+            self.notify(CameraEvent.CAMERA_CONNECT_FAILED)
             raise CameraConnectionError(f"failed to connect: {ee}")
 
         self.log("camera connected.")
-        self._notify(CameraEvent.CAMERA_CONNECTED)
+        self.notify(CameraEvent.CAMERA_CONNECTED)
 
         return self
 
-    def _notify(
+    def notify(
         self,
         event: CameraEvent,
-        extra_payload: Optional[dict[str, Any]] = None,
+        extra_payload: Optional[Dict[str, Any]] = None,
     ):
         """Notifies an event."""
 
@@ -636,7 +652,7 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
 
         self.camera_system.notifier.notify(event, payload)
 
-    def _get_basic_payload(self) -> dict[str, Any]:
+    def _get_basic_payload(self) -> Dict[str, Any]:
         """Returns a dictionary with basic payload for notifying events."""
 
         return {"uid": self.uid, "name": self.name, "camera": self}
@@ -655,7 +671,7 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
         """Returns the cached status. Equivalent to ``get_status(update=False)``."""
         return self.get_status(update=False)
 
-    def get_status(self, update: bool = False) -> dict[str, Any]:
+    def get_status(self, update: bool = False) -> Dict[str, Any]:
         """Returns a dictionary with the camera status values.
 
         Parameters
@@ -670,7 +686,7 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
 
         return self._status
 
-    def _status_internal(self) -> dict[str, Any]:
+    def _status_internal(self) -> Dict[str, Any]:
         """Gets a dictionary with the status of the camera.
 
         This method is intended to be overridden by the specific camera.
@@ -693,6 +709,7 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
         fits_model: Optional[FITSModel] = None,
         filename: Optional[str] = None,
         write: bool = False,
+        postprocess: bool = True,
         **kwargs,
     ) -> Exposure:
         """Exposes the camera.
@@ -727,8 +744,12 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
             `.ImageNamer`.
         write
             If `True`, writes the image to disk immediately.
+        postprocess
+            Whether to run the post-process stage. This argument is ignored if the
+            subclass of `.BaseCamera` does not override ``_post_process_internal``.
         kwargs
-            Other keyword arguments to pass to the internal expose method.
+            Other keyword arguments to pass to the internal expose and post-process
+            methods.
 
         Returns
         -------
@@ -747,45 +768,73 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
             warnings.warn("setting exposure time for bias to 0", ExposureWarning)
             exptime = 0.0
 
-        exposures = []
+        exposures: List[Exposure] = []
 
-        for __ in range(stack):
+        notif_payload = {
+            "exptime": exptime,
+            "remaining_time": exptime,
+            "image_type": image_type,
+            "n_stack": stack,
+            "current_stack": 1,
+        }
+
+        for idx in range(stack):
+
+            notif_payload.update({"n_stack": idx + 1})
 
             exposure = Exposure(self, fits_model=fits_model)
             exposure.image_type = image_type
             exposure.exptime = exptime
 
+            self.notify(CameraEvent.EXPOSURE_INTEGRATING, notif_payload)
             try:
                 await self._expose_internal(exposure, **kwargs)
             except ExposureError as e:
-                self._notify(CameraEvent.EXPOSURE_FAILED, {"error": str(e)})
+                self.notify(CameraEvent.EXPOSURE_FAILED, {"error": str(e)})
                 raise
 
             if exposure.data is None:
                 error = "data was not taken."
-                self._notify(CameraEvent.EXPOSURE_FAILED, {"error": error})
+                self.notify(CameraEvent.EXPOSURE_FAILED, {"error": error})
                 raise ExposureError(error)
 
             exposures.append(exposure)
 
         if len(exposures) > 1:
-            data = [exp.data for exp in exposures]
+            data = [cast(numpy.ndarray, exp.data) for exp in exposures]
             stacked_data = stack_function(numpy.stack(data), axis=0)
             exposures[0].data = stacked_data
 
         exposure = exposures[0]
         exposure.exptime_n = exptime * stack
         exposure.stack = stack
-        exposure.stack_function = stack_function.__name__
+        exposure.stack_function = stack_function
 
-        exposure.filename = filename or self.image_namer(self)
+        if exposure.filename is not None:
+            if filename:
+                exposure.filename = str(filename)
+        else:
+            exposure.filename = str(filename or self.image_namer(self))
 
-        self._notify(CameraEvent.EXPOSURE_READ)
+        notif_payload = {
+            "exptime_total": exposure.exptime_n,
+            "image_type": image_type,
+            "stack": stack,
+        }
+        self.notify(CameraEvent.EXPOSURE_DONE, notif_payload)
+
+        if postprocess:
+            try:
+                exposure = await self._post_process_internal(exposure, **kwargs)
+            except ExposureError:
+                self.notify(CameraEvent.EXPOSURE_POST_PROCESS_FAILED)
+                raise
 
         if write:
-            self._notify(CameraEvent.EXPOSURE_WRITING)
+            filename = os.path.realpath(str(exposure.filename))
+            self.notify(CameraEvent.EXPOSURE_WRITING, {"filename": filename})
             await exposure.write()
-            self._notify(CameraEvent.EXPOSURE_WRITTEN)
+            self.notify(CameraEvent.EXPOSURE_WRITTEN, {"filename": filename})
 
         return exposure
 
@@ -821,17 +870,30 @@ class BaseCamera(LoggerMixIn, metaclass=abc.ABCMeta):
 
         raise NotImplementedError
 
+    async def _post_process_internal(self, exposure: Exposure, **kwargs) -> Exposure:
+        """Performs post-process on an exposure.
+
+        This method can be overridden to perform additional processing on the already
+        read exposure. It's called by `.expose` after `._expose_internal` is complete
+        but before the image is returned to the user or written to disk.
+
+        The user is responsible for issuing any events. If an `.ExposureError` is
+        raised, the error will be propagated and the exposure will not be returned.
+        """
+
+        return exposure
+
     async def disconnect(self) -> bool:
         """Shuts down the camera."""
 
         try:
             await self._disconnect_internal()
         except CameraConnectionError as ee:
-            self._notify(CameraEvent.CAMERA_DISCONNECT_FAILED)
+            self.notify(CameraEvent.CAMERA_DISCONNECT_FAILED)
             raise CameraConnectionError(f"failed to disconnect: {ee}")
 
         self.log("camera has been disconnected.")
-        self._notify(CameraEvent.CAMERA_DISCONNECTED)
+        self.notify(CameraEvent.CAMERA_DISCONNECTED)
 
         return True
 
