@@ -13,19 +13,19 @@ import functools
 import os
 import pathlib
 import re
-import warnings
 
-from typing import Callable, Optional
+from typing import Callable, List, Optional, Tuple, Union
 
 import astropy
-import astropy.io.fits
 import astropy.time
+import astropy.wcs
 import numpy
+from astropy.io.fits import BinTableHDU, HDUList, ImageHDU
 
 import basecam.camera
 import basecam.models
 
-from .exceptions import ExposureError, ExposureWarning
+from .exceptions import ExposureError
 
 
 __all__ = ["Exposure", "ImageNamer"]
@@ -49,6 +49,8 @@ class Exposure(object):
     fits_model
         The `model <.FITSModel>` to create the FITS image. If `None`, a single
         extension with a basic header will be used.
+    wcs
+        The WCS object describing the astrometry of this exposure.
 
     Attributes
     ----------
@@ -66,6 +68,8 @@ class Exposure(object):
         Name of the function used for stacking.
     filename
         The path where to write the image.
+    wcs : ~astropy.wcs.WCS
+        The WCS object describing the astrometry of this exposure.
     """
 
     def __init__(
@@ -74,11 +78,12 @@ class Exposure(object):
         filename: Optional[str] = None,
         data: Optional[numpy.ndarray] = None,
         fits_model: Optional[basecam.models.FITSModel] = None,
+        wcs: Optional[astropy.wcs.WCS] = None,
     ):
 
         self.camera = camera
         self.data = data
-        self.fits_model = fits_model or basecam.models.FITSModel()
+        self.fits_model = fits_model.copy() if fits_model else None
         self.filename = filename
 
         self._obstime: astropy.time.Time = astropy.time.Time.now()
@@ -87,9 +92,12 @@ class Exposure(object):
         self.exptime_n: Optional[float] = None
         self.stack: int = 1
         self.stack_function: Optional[Callable[..., numpy.ndarray]] = None
-
         self.image_type: Optional[str] = None
 
+        self._extra_hdus: List[Tuple[Union[BinTableHDU, ImageHDU], Optional[int]]] = []
+
+        self.wcs = wcs
+
     @property
     def obstime(self) -> astropy.time.Time:
         """The time at the beginning of the observation.
@@ -110,7 +118,27 @@ class Exposure(object):
         else:
             raise ExposureError(f"invalid obstime {value}")
 
-    def to_hdu(self, context={}) -> astropy.io.fits.HDUList:
+    def add_hdu(self, hdu: Union[BinTableHDU, ImageHDU], index: Optional[int] = None):
+        """Adds an HDU to the list of extensions.
+
+        Parameters
+        ----------
+        hdu
+            The `~astropy.io.fits.BinTableHDU` or `~astropy.io.fits.ImageHDU` HDU to
+            append.
+        index
+            The index where the extension will be added. Extra HDUs are inserted in
+            order after the FITS model has been generated. ``index=None`` appends the
+            new HDU at the end of the list. Note that ``astropy.io.fits`` may change
+            the final order of the extensions to ensure that a primary HDU remains
+            as the first HDU.
+        """
+
+        assert isinstance(hdu, (BinTableHDU, ImageHDU)), "invalid HDU type"
+
+        self._extra_hdus.append((hdu, index))
+
+    def to_hdu(self, context={}) -> HDUList:
         """Return an `~astropy.io.fits.HDUList` for the image.
 
         Parameters
@@ -125,15 +153,16 @@ class Exposure(object):
             for this exposure.
         """
 
-        fits_model = self.fits_model
+        fits_model = self.fits_model or basecam.models.FITSModel()
 
-        if not fits_model:
-            warnings.warn(
-                "No FITS model defined. Reverting to default one.", ExposureWarning
-            )
-            fits_model = basecam.models.FITSModel()
+        hdulist = fits_model.to_hdu(self, context=context)
+        for hdu, index in self._extra_hdus:
+            if index:
+                hdulist.insert(index, hdu)
+            else:
+                hdulist.append(hdu)
 
-        return fits_model.to_hdu(self, context=context)
+        return hdulist
 
     async def write(
         self,
@@ -141,7 +170,7 @@ class Exposure(object):
         context={},
         overwrite=False,
         checksum=True,
-    ) -> astropy.io.fits.HDUList:
+    ) -> HDUList:
         """Writes the image to disk.
 
         Parameters
@@ -229,7 +258,7 @@ class ImageNamer(object):
         self._basename: str
         self.basename = basename
 
-        self.dirname: pathlib.Path | str = pathlib.Path(dirname)
+        self.dirname: Union[pathlib.Path, str] = pathlib.Path(dirname)
         self.overwrite: bool = overwrite
 
         self._last_num = 0
@@ -246,7 +275,7 @@ class ImageNamer(object):
         # We want to expand everything except the num first so we "double-escape" it.
         self._basename = re.sub(r"(\{num.+\})", r"{\1}", value)
 
-    def _eval_dirname(self) -> pathlib.Path:
+    def get_dirname(self) -> pathlib.Path:
         """Returns the evaluated dirname."""
 
         date = astropy.time.Time.now()
@@ -263,7 +292,7 @@ class ImageNamer(object):
 
         regex = re.compile(re.sub(r"\{num.+\}", "(?P<num>[0-9]*?)", basename))
 
-        dirname = self._eval_dirname()
+        dirname = self.get_dirname()
         all_files = list(map(str, dirname.glob("*")))
 
         match_files = list(filter(regex.search, all_files))
@@ -277,6 +306,7 @@ class ImageNamer(object):
     def __call__(
         self,
         camera: Optional[basecam.camera.BaseCamera] = None,
+        update_num: bool = True,
     ) -> pathlib.Path:
 
         camera = camera or self.camera
@@ -287,7 +317,9 @@ class ImageNamer(object):
             expanded_basename = self.basename.format()
 
         num = self._get_num(expanded_basename)
-        path = self._eval_dirname() / expanded_basename.format(num=num)
-        self._last_num = num
+        path = self.get_dirname() / expanded_basename.format(num=num)
+
+        if update_num:
+            self._last_num = num
 
         return path
